@@ -5,23 +5,21 @@
 
 #include "lib/regex.h"
 #include "lib/regex_impl.h"
-#include "src/debug/debug.h"
 #include "src/dfa/closure_leftmost.h"
 #include "src/dfa/tag_history.h"
 #include "src/nfa/nfa.h"
 #include "src/regexp/rule.h"
+#include "src/util/check.h"
 #include "src/util/range.h"
 
-
 namespace re2c {
 namespace libre2c {
 
-static void reach_on_symbol(lsimctx_t &, uint32_t);
+static void reach_on_symbol(lsimctx_t&, uint32_t);
 
-int regexec_nfa_leftmost(const regex_t *preg, const char *string
-    , size_t nmatch, regmatch_t pmatch[], int)
-{
-    lsimctx_t &ctx = *static_cast<lsimctx_t*>(preg->simctx);
+int regexec_nfa_leftmost(
+        const regex_t* preg, const char* string, size_t nmatch, regmatch_t pmatch[], int) {
+    lsimctx_t& ctx = *static_cast<lsimctx_t*>(preg->simctx);
     init(ctx, string);
 
     // root state can be non-core, so we pass zero as origin to avoid checks
@@ -36,9 +34,9 @@ int regexec_nfa_leftmost(const regex_t *preg, const char *string
     }
 
     for (cconfiter_t i = ctx.state.begin(), e = ctx.state.end(); i != e; ++i) {
-        nfa_state_t *s = i->state;
+        TnfaState* s = i->state;
         s->clos = NOCLOS;
-        if (s->type == nfa_state_t::FIN) {
+        if (s->kind == TnfaState::Kind::FIN) {
             update_offsets(ctx, *i, NONCORE);
         }
     }
@@ -53,30 +51,28 @@ int regexec_nfa_leftmost(const regex_t *preg, const char *string
     return 0;
 }
 
-void reach_on_symbol(lsimctx_t &ctx, uint32_t sym)
-{
-    const confset_t &state = ctx.state;
-    confset_t &reach = ctx.reach;
-    DASSERT(reach.empty());
+void reach_on_symbol(lsimctx_t& ctx, uint32_t sym) {
+    const confset_t& state = ctx.state;
+    confset_t& reach = ctx.reach;
+    DCHECK(reach.empty());
 
     // in reverse, so that future closure DFS has states in stack order
     uint32_t j = 0;
     for (rcconfiter_t i = state.rbegin(), e = state.rend(); i != e; ++i) {
-        nfa_state_t *s = i->state;
+        TnfaState* s = i->state;
         s->clos = NOCLOS;
 
-        if (s->type == nfa_state_t::RAN) {
-            for (const Range *r = s->ran.ran; r; r = r->next()) {
+        if (s->kind == TnfaState::Kind::RAN) {
+            for (const Range* r = s->ran; r; r = r->next()) {
                 if (r->lower() <= sym && sym < r->upper()) {
-                    conf_t c(s->ran.out, j, HROOT);
+                    conf_t c(s->out1, j, HROOT);
                     reach.push_back(c);
                     update_offsets(ctx, *i, j);
                     ++j;
                     break;
                 }
             }
-        }
-        else if (s->type == nfa_state_t::FIN) {
+        } else if (s->kind == TnfaState::Kind::FIN) {
             update_offsets(ctx, *i, NONCORE);
         }
     }
@@ -88,4 +84,3 @@ void reach_on_symbol(lsimctx_t &ctx, uint32_t sym)
 
 } // namespace libre2
 } // namespace re2c
-
