@@ -9,6 +9,7 @@
 #include "src/msg/location.h"
 #include "src/util/allocator.h"
 #include "src/util/check.h"
+#include "src/util/containers.h"
 #include "src/util/string_utils.h"
 
 namespace re2c {
@@ -33,13 +34,7 @@ inline Label* new_label(OutAllocator& alc, uint32_t index) {
     return l;
 }
 
-template<typename T>
-struct code_list_t {
-    T* head;
-    T** ptail;
-};
-
-using CodeList = code_list_t<Code>;
+using CodeList = list_t<Code>;
 
 struct CodeJump {
     const State* to;
@@ -57,11 +52,13 @@ struct CodeBmState {
     CodeBmState* next;
 };
 
-using CodeBmStates = code_list_t<CodeBmState>;
+using CodeBmStates = list_t<CodeBmState>;
 
 struct CodeBitmap {
     CodeBmStates* states;
     uint32_t nchars;
+    uint32_t nelems;
+    const char** elems;
     bool used;
 };
 
@@ -88,60 +85,52 @@ struct CodeGoSw {
     uint32_t ncases;
 };
 
-struct CodeCmp {
-    const char* cmp;
-    uint32_t val;
-};
-
-// binary if
-struct CodeGoIfB {
-    const CodeCmp* cond;
-    CodeGoIf* gothen;
-    CodeGoIf* goelse;
-};
-
-// linear if
-struct CodeGoIfL {
-    struct Branch {
-        const CodeCmp* cond;
-        CodeJump jump;
-    };
-
-    size_t nbranches;
-    Branch* branches;
-    State* def;
-};
-
-struct CodeGoIf {
-    enum class Kind: uint32_t {
-        BINARY,
-        LINEAR
-    };
-
-    Kind kind;
-    union {
-        CodeGoIfB* goifb;
-        CodeGoIfL* goifl;
-    };
-};
-
 struct CodeGoSwIf {
-    enum class Kind: uint32_t {
-        SWITCH,
-        IF
-    };
-
-    Kind kind;
     union {
         CodeGoSw* gosw;
         CodeGoIf* goif;
     };
+    enum class Kind: uint32_t {
+        SWITCH,
+        IF
+    } kind;
 };
 
-struct CodeGoBm {
-    const CodeBmState* bitmap;
-    CodeGoSwIf* hgo;
-    CodeGoSwIf* lgo;
+// Binary IF (used for many branches).
+struct CodeGoIfB {
+    const char* cond;
+    CodeGoIf* gothen;
+    CodeGoIf* goelse;
+};
+
+struct CodeGoBranch {
+    const char* cond;
+    union {
+        CodeJump jump;
+        CodeGoSwIf* swif;
+    };
+    enum class Kind: uint32_t {
+        JUMP,
+        SWIF
+    } kind;
+};
+
+// Linear IF (used for few branches).
+struct CodeGoIfL {
+    size_t nbranches;
+    CodeGoBranch* branches;
+    State* def;
+};
+
+struct CodeGoIf {
+    union {
+        CodeGoIfB* goifb;
+        CodeGoIfL* goifl;
+    };
+    enum class Kind: uint32_t {
+        BINARY,
+        LINEAR
+    } kind;
 };
 
 struct CodeGoCpTable {
@@ -159,7 +148,7 @@ struct CodeGo {
     enum class Kind: uint32_t {
         EMPTY,
         SWITCH_IF,
-        BITMAP,
+        LINEAR_IF,
         CPGOTO,
         DOT
     };
@@ -171,20 +160,20 @@ struct CodeGo {
     bool skip;
     union {
         CodeGoSwIf* goswif;
-        CodeGoBm* gobm;
         CodeGoCp* gocp;
         CodeGoSw* godot;
+        CodeGoIfL* goifl;
     };
 };
 
-struct CodeIfTE {
-    const char* if_cond;
-    const char* else_cond;
-    CodeList* if_code;
-    CodeList* else_code;
-    bool oneline;
+struct CodeBranch {
+    const char* cond;
+    CodeList* code;
+    CodeBranch* next;
 };
 
+using CodeBranches = list_t<CodeBranch>;
+
 struct CodeCase {
     enum class Kind: uint32_t {
         RANGES,
@@ -202,7 +191,7 @@ struct CodeCase {
     CodeCase* next;
 };
 
-using CodeCases = code_list_t<CodeCase>;
+using CodeCases = list_t<CodeCase>;
 
 struct CodeSwitch {
     const char* expr;
@@ -222,15 +211,22 @@ struct CodeBlock {
 
 struct CodeVar {
     VarType type;
+    bool is_default;
     const char* name;
     const char* init;
 };
 
+// TODO: merge `BlockNameList` and `CodeExpr`?
 struct BlockNameList {
     const char* name;
     BlockNameList* next;
 };
 
+struct CodeAssign {
+    const char* lhs;
+    const char* rhs;
+};
+
 struct CodeFmt {
     BlockNameList* block_names;
     const char* format;
@@ -242,17 +238,33 @@ struct CodeRaw {
     size_t size;
 };
 
+struct CodeParam {
+    const char* name;
+    const char* type;
+    CodeParam* next;
+};
+
+using CodeParams = list_t<CodeParam>;
+
 struct CodeArg {
     const char* arg;
     CodeArg* next;
 };
 
-using CodeArgs = code_list_t<CodeArg>;
+using CodeArgs = list_t<CodeArg>;
 
-struct CodeFunc {
-    CodeArgs* args;
+struct CodeFnDef {
     const char* name;
-    const char* semi;
+    const char* type;
+    const CodeParams* params; // shared between different function definitions
+    CodeList* body;
+};
+
+struct CodeFnCall {
+    const char* name;
+    const char* retval;
+    const CodeArgs* args; // shared between different function calls
+    bool tailcall;
 };
 
 struct CodeLabel {
@@ -262,65 +274,78 @@ struct CodeLabel {
     } kind;
 
     union {
-        Label* nlabel;
+        const Label* nlabel;
         const char* slabel;
     };
 };
 
+struct CodeArray {
+    const char* name;
+    const char* type;
+    const char** elems;
+    size_t size;
+    bool local;
+    bool tabulate;
+};
+
+struct CodeEnum {
+    const char* type;
+    size_t size;
+    const char** elem_ids;
+    const uint32_t* elem_nums;
+};
+
+struct CodeDebug {
+    uint32_t state;
+};
+
+struct CodeTag {
+    const char* tag1;
+    const char* tag2;
+    int32_t dist;
+};
+
 struct Code {
     union {
         const char* text;
+        const char* target;
+        const char* cond;
+        const char* state;
         BlockNameList* block_names;
-        CodeIfTE ifte;
+        CodeBranches* ifte;
         CodeSwitch swch;
         CodeBlock block;
-        CodeFunc func;
+        CodeFnDef fndef;
+        CodeFnCall fncall;
         CodeRaw raw;
         CodeVar var;
+        CodeArray array;
+        CodeEnum enumr;
         CodeFmt fmt;
+        CodeAssign assign;
         CodeLabel label;
+        CodeDebug debug;
+        CodeTag tag;
         CodeList* loop;
+        CodeList* rfuncs;
         loc_t loc;
+        size_t accept;
     };
 
     CodeKind kind;
     Code* next;
 };
 
-template<typename T>
-inline code_list_t<T>* new_code_list(OutAllocator& alc) {
-    code_list_t<T>* x = alc.alloct<code_list_t<T> >(1);
-    x->head = nullptr;
-    x->ptail = &x->head;
-    return x;
-}
-
-template<typename T>
-inline void append(code_list_t<T>* list, T* elem) {
-    if (elem != nullptr) {
-        *list->ptail = elem;
-        list->ptail = &elem->next;
-    }
-}
-
-template<typename T>
-inline void prepend(code_list_t<T>* list, T* elem) {
-    if (elem != nullptr) {
-        if (!list->head) {
-            list->ptail = &elem->next;
-        }
-        elem->next = list->head;
-        list->head = elem;
-    }
-}
-
-template<typename T>
-inline void append(code_list_t<T>* list1, code_list_t<T>* list2) {
-    if (list2 && list2->head) {
-        *list1->ptail = list2->head;
-        for (; *(list1->ptail); list1->ptail = &(*list1->ptail)->next);
-    }
-}
+// YYFN name, type, attributes and params/args (they are common for all function definitions/calls,
+// so they can be precomputed for each block options and once for whole-program options).
+struct CodeFnCommon {
+    const char* name;
+    const char* type;
+    CodeParams* params = nullptr;
+    CodeParams* params_yych = nullptr; // same as `params`, but with `yych` at the end
+    CodeArgs* args = nullptr;
+    CodeArgs* args_yych = nullptr; // same as `args`, but with `yych` at the end
+};
 
 inline Code* new_code(OutAllocator& alc, CodeKind kind) {
     Code* x = alc.alloct<Code>(1);
@@ -335,6 +360,18 @@ inline Code* code_stmt(OutAllocator& alc, const char* text) {
     return x;
 }
 
+inline Code* code_goto(OutAllocator& alc, const char* target) {
+    Code* x = new_code(alc, CodeKind::GOTO);
+    x->target = target;
+    return x;
+}
+
+inline Code* code_continue(OutAllocator& alc, const char* target) {
+    Code* x = new_code(alc, CodeKind::CONTINUE);
+    x->target = target;
+    return x;
+}
+
 inline Code* code_loop(OutAllocator& alc, CodeList* loop) {
     Code* x = new_code(alc, CodeKind::LOOP);
     x->loop = loop;
@@ -370,7 +407,7 @@ inline Code* code_newline(OutAllocator& alc) {
     return code_textraw(alc, "");
 }
 
-inline Code* code_nlabel(OutAllocator& alc, Label* label) {
+inline Code* code_nlabel(OutAllocator& alc, const Label* label) {
     Code* x = new_code(alc, CodeKind::LABEL);
     x->label.kind = CodeLabel::Kind::NLABEL;
     x->label.nlabel = label;
@@ -402,30 +439,46 @@ inline Code* code_state_goto(OutAllocator& alc, BlockNameList* blocks) {
     return x;
 }
 
-inline Code* code_line_info_input(OutAllocator& alc, Lang lang, const loc_t& loc) {
-    // Rust has no line directives.
-    if (lang == Lang::RUST) return nullptr;
-
+inline Code* code_line_info_input(OutAllocator& alc, const loc_t& loc) {
     Code* x = new_code(alc, CodeKind::LINE_INFO_INPUT);
     x->loc = loc;
     return x;
 }
 
-inline Code* code_var(OutAllocator& alc, VarType type, const std::string& name, const char* init) {
+inline Code* code_var(
+        OutAllocator& alc, VarType type, bool is_default, const char* name, const char* init) {
     Code* x = new_code(alc, CodeKind::VAR);
     x->var.type = type;
+    x->var.is_default = is_default;
     x->var.name = copystr(name, alc);
     x->var.init = init;
     return x;
 }
 
-inline Code* code_line_info_output(OutAllocator& alc, Lang lang) {
-    // Rust has no line directives.
-    if (lang == Lang::RUST) return nullptr;
+inline void init_code_const(Code* x, VarType type, const char* name, const char* init) {
+    x->kind = CodeKind::CONST;
+    x->var.type = type;
+    x->var.name = name;
+    x->var.init = init;
+}
 
+inline Code* code_assign(OutAllocator& alc, const char* lhs, const char* rhs) {
+    Code* x = new_code(alc, CodeKind::ASSIGN);
+    x->assign.lhs = lhs;
+    x->assign.rhs = rhs;
+    return x;
+}
+
+inline Code* code_line_info_output(OutAllocator& alc) {
     return new_code(alc, CodeKind::LINE_INFO_OUTPUT);
 }
 
+inline Code* code_debug(OutAllocator& alc, uint32_t state) {
+    Code* x = new_code(alc, CodeKind::DEBUG);
+    x->debug.state = state;
+    return x;
+}
+
 inline Code* code_skip(OutAllocator& alc) {
     return new_code(alc, CodeKind::SKIP);
 }
@@ -438,6 +491,79 @@ inline Code* code_backup(OutAllocator& alc) {
     return new_code(alc, CodeKind::BACKUP);
 }
 
+inline Code* code_backupctx(OutAllocator& alc) {
+    return new_code(alc, CodeKind::BACKUPCTX);
+}
+
+inline Code* code_restore(OutAllocator& alc) {
+    return new_code(alc, CodeKind::RESTORE);
+}
+
+inline Code* code_restore_ctx(OutAllocator& alc) {
+    return new_code(alc, CodeKind::RESTORECTX);
+}
+
+inline Code* code_restore_tag(OutAllocator& alc, const char* tag) {
+    Code* x = new_code(alc, CodeKind::RESTORETAG);
+    x->tag.tag1 = tag;
+    x->tag.tag2 = nullptr;
+    x->tag.dist = 0;
+    return x;
+}
+
+inline Code* code_shift(OutAllocator& alc, int32_t dist) {
+    Code* x = new_code(alc, CodeKind::SHIFT);
+    x->tag.tag1 = nullptr;
+    x->tag.tag2 = nullptr;
+    x->tag.dist = dist;
+    return x;
+}
+
+inline Code* code_shift_tag(
+        OutAllocator& alc, const char* tag, const char* neg, int32_t dist, bool is_mtag) {
+    Code* x = new_code(alc, is_mtag ? CodeKind::SHIFTMTAG : CodeKind::SHIFTSTAG);
+    x->tag.tag1 = tag;
+    x->tag.tag2 = neg;
+    x->tag.dist = dist;
+    return x;
+}
+
+inline Code* code_set_tag(OutAllocator& alc, const char* tag, bool is_mtag, bool is_negative) {
+    Code* x = new_code(alc, is_mtag
+        ? (is_negative ? CodeKind::MTAGN : CodeKind::MTAGP)
+        : (is_negative ? CodeKind::STAGN : CodeKind::STAGP));
+    x->tag.tag1 = tag;
+    x->tag.tag2 = nullptr;
+    x->tag.dist = 0;
+    return x;
+}
+
+inline Code* code_copy_tag(OutAllocator& alc, const char* lhs, const char* rhs, bool is_mtag) {
+    Code* x = new_code(alc, is_mtag ? CodeKind::COPYMTAG : CodeKind::COPYSTAG);
+    x->tag.tag1 = lhs;
+    x->tag.tag2 = rhs;
+    x->tag.dist = 0;
+    return x;
+}
+
+inline Code* code_set_accept(OutAllocator& alc, size_t accept) {
+    Code* x = new_code(alc, CodeKind::SETACCEPT);
+    x->accept = accept;
+    return x;
+}
+
+inline Code* code_set_cond(OutAllocator& alc, const char* cond) {
+    Code* x = new_code(alc, CodeKind::SETCOND);
+    x->cond = cond;
+    return x;
+}
+
+inline Code* code_set_state(OutAllocator& alc, const char* state) {
+    Code* x = new_code(alc, CodeKind::SETSTATE);
+    x->state = state;
+    return x;
+}
+
 inline Code* code_dfas(OutAllocator& alc) {
     return new_code(alc, CodeKind::DFAS);
 }
@@ -449,17 +575,30 @@ inline Code* code_block(OutAllocator& alc, CodeList* stmts, CodeBlock::Kind kind
     return x;
 }
 
-inline Code* code_if_then_else(OutAllocator& alc,
-                               const char* if_cond,
-                               CodeList* if_code,
-                               CodeList* else_code,
-                               bool oneline = true) {
+inline CodeBranch* code_branch(OutAllocator& alc, const char* cond, CodeList* code) {
+    CodeBranch* x = alc.alloct<CodeBranch>(1);
+    x->cond = cond;
+    x->code = code;
+    x->next = nullptr;
+    return x;
+}
+
+inline CodeBranches* code_branches(OutAllocator& alc) {
+    return new_list<CodeBranch>(alc);
+}
+
+inline Code* code_if_then_else(OutAllocator& alc) {
     Code* x = new_code(alc, CodeKind::IF_THEN_ELSE);
-    x->ifte.if_cond = if_cond;
-    x->ifte.else_cond = nullptr;
-    x->ifte.if_code = if_code;
-    x->ifte.else_code = else_code;
-    x->ifte.oneline = oneline;
+    x->ifte = code_branches(alc);
+    return x;
+}
+
+inline Code* code_if_then_else(
+        OutAllocator& alc, const char* if_cond, CodeList* if_code, CodeList* else_code) {
+    Code* x = new_code(alc, CodeKind::IF_THEN_ELSE);
+    x->ifte = code_branches(alc);
+    append(x->ifte, code_branch(alc, if_cond, if_code));
+    if (else_code) append(x->ifte, code_branch(alc, nullptr, else_code));
     return x;
 }
 
@@ -468,8 +607,10 @@ inline Code* code_if_then_elif(OutAllocator& alc,
                                CodeList* if_code,
                                const char* else_cond,
                                CodeList* else_code) {
-    Code* x = code_if_then_else(alc, if_cond, if_code, else_code, false);
-    x->ifte.else_cond = else_cond;
+    Code* x = new_code(alc, CodeKind::IF_THEN_ELSE);
+    x->ifte = code_branches(alc);
+    append(x->ifte, code_branch(alc, if_cond, if_code));
+    append(x->ifte, code_branch(alc, else_cond, else_code));
     return x;
 }
 
@@ -497,15 +638,13 @@ inline CodeCase* code_case_string(OutAllocator& alc, CodeList* body, const char*
     return x;
 }
 
-inline CodeCase* code_case_ranges(
-    OutAllocator& alc, CodeList* body, const CodeRanges* ranges) {
+inline CodeCase* code_case_ranges(OutAllocator& alc, CodeList* body, const CodeRanges* ranges) {
     CodeCase* x = code_case(alc, body, CodeCase::Kind::RANGES);
     x->ranges = ranges;
     return x;
 }
 
-inline CodeRanges* code_ranges(
-    OutAllocator& alc, VarType type, int64_t* start, int64_t* end) {
+inline CodeRanges* code_ranges(OutAllocator& alc, VarType type, int64_t* start, int64_t* end) {
     CodeRanges* x = alc.alloct<CodeRanges>(1);
     x->size = static_cast<uint32_t>(end - start) / 2;
     x->type = type;
@@ -514,7 +653,19 @@ inline CodeRanges* code_ranges(
 }
 
 inline CodeCases* code_cases(OutAllocator& alc) {
-    return new_code_list<CodeCase>(alc);
+    return new_list<CodeCase>(alc);
+}
+
+inline CodeParam* code_param(OutAllocator& alc, const char* name, const char* type) {
+    CodeParam* x = alc.alloct<CodeParam>(1);
+    x->name = name;
+    x->type = type;
+    x->next = nullptr;
+    return x;
+}
+
+inline CodeParams* code_params(OutAllocator& alc) {
+    return new_list<CodeParam>(alc);
 }
 
 inline CodeArg* code_arg(OutAllocator& alc, const char* arg) {
@@ -525,23 +676,52 @@ inline CodeArg* code_arg(OutAllocator& alc, const char* arg) {
 }
 
 inline CodeArgs* code_args(OutAllocator& alc) {
-    return new_code_list<CodeArg>(alc);
+    return new_list<CodeArg>(alc);
 }
 
-inline Code* code_func(OutAllocator& alc, const char* name, CodeArgs* args, const char* semi) {
-    Code* x = new_code(alc, CodeKind::FUNC);
-    x->func.args = args;
-    x->func.name = name;
-    x->func.semi = semi;
+inline void init_code_fndef(
+        Code* x, const char* name, const char* type, CodeParams* params, CodeList* body) {
+    x->kind = CodeKind::FNDEF;
+    x->fndef.name = name;
+    x->fndef.type = type;
+    x->fndef.params = params;
+    x->fndef.body = body;
+}
+
+inline Code* code_fndef(
+        OutAllocator& alc, const char* name, const char* type, CodeParams* params, CodeList* body) {
+    Code* x = new_code(alc, CodeKind::FNDEF);
+    x->fndef.name = name;
+    x->fndef.type = type;
+    x->fndef.params = params;
+    x->fndef.body = body;
     return x;
 }
 
-inline Code* code_fdecl(OutAllocator& alc, const char* name, CodeArgs* args) {
-    return code_func(alc, name, args, "");
+inline Code* code_fncall(
+        OutAllocator& alc,
+        const char* name,
+        const char* retval,
+        CodeArgs* args,
+        bool tailcall = false) {
+    Code* x = new_code(alc, CodeKind::FNCALL);
+    x->fncall.name = name;
+    x->fncall.retval = retval;
+    x->fncall.args = args;
+    x->fncall.tailcall = tailcall;
+    return x;
 }
 
-inline Code* code_fcall(OutAllocator& alc, const char* name, CodeArgs* args, const char* semi) {
-    return code_func(alc, name, args, semi);
+inline Code* code_tailcall(OutAllocator& alc, const char* name, CodeArgs* args, bool have_retval) {
+    // tailcall returns immediately so the `retval` string won't be used
+    const char* retval = have_retval ? "<unused-retval>" : nullptr;
+    return code_fncall(alc, name, retval, args, true);
+}
+
+inline Code* code_recursive_functions(OutAllocator& alc, CodeList* fndefs) {
+    Code* x = new_code(alc, CodeKind::REC_FUNCS);
+    x->rfuncs = fndefs;
+    return x;
 }
 
 inline Code* code_switch(OutAllocator& alc, const char* expr, CodeCases* cases) {
@@ -552,7 +732,7 @@ inline Code* code_switch(OutAllocator& alc, const char* expr, CodeCases* cases)
 }
 
 inline CodeList* code_list(OutAllocator& alc) {
-    return new_code_list<Code>(alc);
+    return new_list<Code>(alc);
 }
 
 inline CodeBmState* code_bmstate(OutAllocator& alc, const CodeGo* go, const State* s) {
@@ -567,19 +747,45 @@ inline CodeBmState* code_bmstate(OutAllocator& alc, const CodeGo* go, const Stat
 
 inline CodeBitmap* code_bitmap(OutAllocator& alc, uint32_t nchars) {
     CodeBitmap* x = alc.alloct<CodeBitmap>(1);
-    x->states = new_code_list<CodeBmState>(alc);
+    x->states = new_list<CodeBmState>(alc);
     x->nchars = nchars;
+    x->nelems = 0;
+    x->elems = nullptr;
     x->used = false;
     return x;
 }
 
-inline CodeCmp* code_cmp(OutAllocator& alc, const char* cmp, uint32_t val) {
-    CodeCmp* x = alc.alloct<CodeCmp>(1);
-    x->cmp = cmp;
-    x->val = val;
+inline Code* code_array(
+        OutAllocator& alc,
+        const char* name,
+        const char* type,
+        const char** elems,
+        size_t size,
+        bool local,
+        bool tabulate = false) {
+    Code* x = new_code(alc, CodeKind::ARRAY);
+    x->array.name = name;
+    x->array.type = type;
+    x->array.elems = elems;
+    x->array.size = size;
+    x->array.local = local;
+    x->array.tabulate = tabulate;
     return x;
 }
 
+inline void init_code_enum(
+        Code* x, const char* type, size_t size, const char** elem_ids, uint32_t* elem_nums) {
+    x->kind = CodeKind::ENUM;
+    x->enumr.type = type;
+    x->enumr.size = size;
+    x->enumr.elem_ids = elem_ids;
+    x->enumr.elem_nums = elem_nums;
+}
+
+inline Code* code_fingerprint(OutAllocator& alc) {
+    return new_code(alc, CodeKind::FINGERPRINT);
+}
+
 } // namespace re2c
 
 #endif // _RE2C_CODEGEN_CODE_
